from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from .config import get_config
from .console import print_and_raise, print_warn

if TYPE_CHECKING:
    from .cache import CacheData


def validate_search_paths() -> list[Path]:
    """Return existing configured search paths.

    Raises:
        typer.Exit: If no valid search paths are configured.

    Returns:
        list[Path]: List of validated existing search paths.
    """
    search_paths = get_config()["search_paths"]
    validated: list[Path] = []

    for search_path in search_paths:
        p = Path(search_path)

        if not p.exists():
            print_warn(f"Configured search path {search_path} does not exist.")
        else:
            validated.append(p)

    if not validated:
        print_and_raise(
            "List of search paths is empty or paths don't exist. "
            "Set search paths with 'mf config set search_paths'."
        )

    return validated


def validate_cache_structure(data: dict) -> CacheData:
    """Validate that cache has the required structure."""
    if not isinstance(data, dict):
        raise ValueError("Cache must be a dictionary.")

    if "timestamp" not in data or "files" not in data:
        raise ValueError("Cache missing required keys.")

    if not isinstance(data["files"], list):
        raise ValueError("Cache 'files' must be a list.")

    return data



import json

from .console import print_and_raise
from .file import FileResult, get_search_cache_file, open_utf8


def save_last_played(result: FileResult):
    """Save which file was played last to the cached search results file.

    Args:
        result (FileResult): File last played.
    """
    with open_utf8(get_search_cache_file()) as f:
        cached = json.load(f)

    last_search_results: list[str] = cached["results"]
    last_played_index = last_search_results.index(str(result))
    cached["last_played_index"] = last_played_index

    with open_utf8(get_search_cache_file(), "w") as f:
        json.dump(cached, f, indent=2)


def get_last_played_index() -> int | None:
    """Get the search result index of the last played file.

    Returns:
        int | None: Index or None if no file was played.
    """
    with open_utf8(get_search_cache_file()) as f:
        cached = json.load(f)

    try:
        return int(cached["last_played_index"])
    except KeyError:
        return None


def get_next() -> FileResult:
    """Get the next file to play.

    Returns:
        FileResult: Next file to play.
    """
    with open_utf8(get_search_cache_file()) as f:
        cached = json.load(f)

    results: list[str] = cached["results"]

    try:
        index_last_played = int(cached["last_played_index"])
    except KeyError:
        # Nothing played yet, start at the beginning
        index_last_played = -1

    try:
        return FileResult.from_string(results[index_last_played + 1])
    except IndexError as e:
        print_and_raise("Last available file already played.", raise_from=e)



import json
from json import JSONDecodeError
from urllib import request
from urllib.error import URLError

from packaging.version import Version

from .utils.console import print_and_raise, print_info, print_ok

__version__ = "0.8.1"


def get_pypi_version() -> Version:
    """Get number of latest version published on PyPI
    (https://pypi.org/pypi/mediafinder).

    Returns:
        Version: Version number.
    """
    url = "https://pypi.org/pypi/mediafinder/json"

    try:
        with request.urlopen(url) as response:
            data = json.loads(response.read().decode())
            return Version(data["info"]["version"])
    except URLError as e:
        print_and_raise(f"Network error checking version: {e}", raise_from=e)
    except JSONDecodeError as e:
        print_and_raise(f"Invalid response from PyPI: {e}", raise_from=e)
    except KeyError as e:
        print_and_raise(f"Unexpected PyPI API response format: {e}", raise_from=e)


def check_version():
    """Check installed version against latest available version of mediafinder."""
    pypi_version = get_pypi_version()
    local_version = Version(__version__)

    if pypi_version > local_version:
        print_info(
            "There's a newer version of mediafinder available "
            f"({local_version} → {pypi_version}). "
            "Use 'uv tool upgrade mediafinder' to upgrade."
        )
    else:
        print_ok(f"You're on the latest version of mediafinder ({local_version}).")



from types import SimpleNamespace

import mf.utils.misc as misc_mod


def test_get_vlc_command_posix(monkeypatch):
    # Force POSIX branch
    monkeypatch.setattr(misc_mod, "os", SimpleNamespace(name="posix"))
    assert misc_mod.get_vlc_command() == "vlc"



import json
from datetime import datetime, timedelta
from typing import TypedDict

from .config import build_config
from .console import print_info, print_ok, print_warn
from .file import FileResults, get_library_cache_file, open_utf8
from .validation import validate_cache_structure

StatList = tuple[
    int,  # st_mode
    int,  # st_ino
    int,  # st_dev
    int,  # st_nlink
    int,  # st_uid
    int,  # st_gid
    int,  # st_size
    int,  # st_atime
    int,  # st_mtime
    int,  # st_ctime
]
FileEntry = tuple[
    str,  # File path
    StatList,
]


class CacheData(TypedDict):
    """Media library cache data structure.

    Contains metadata for all files found during library scanning,
    including file paths and their filesystem stat information.

    Attributes:
        timestamp: ISO format timestamp of when the cache was last rebuilt
        files: List of [file_path, stat_list] pairs where:

            - file_path: Absolute POSIX path to the media file
            - stat_list: os.stat_result as 10-element list containing

              [st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid,
               st_size, st_atime, st_mtime, st_ctime]

    Example:
        {
            "timestamp": "2025-01-02T10:30:00.123456",
            "files": [
                [
                 "/path/to/movie.mkv",
                 [33206, 0, 0, 0, 0, 0, 1234567890, 1640995200, 1640995200, 1640995200]
                ],
                ...
            ]
        }
    """

    timestamp: str
    files: list[FileEntry]


def rebuild_library_cache() -> FileResults:
    """Rebuild the local library cache.

    Builds an mtime-sorted index (descending / newest first) of all media files in the
    configured search paths.

    Returns:
        FileResults: Rebuilt cache.
    """
    from .scan import scan_search_paths

    print_info("Rebuilding cache.")
    results = scan_search_paths(cache_stat=True, show_progress=True)
    results.sort(by_mtime=True)
    cache_data = {
        "timestamp": datetime.now().isoformat(),
        "files": [(result.file.as_posix(), tuple(result.stat)) for result in results],
    }

    with open_utf8(get_library_cache_file(), "w") as f:
        json.dump(cache_data, f, indent=2)

    print_ok("Cache rebuilt.")
    return results


def _load_library_cache(allow_rebuild=True) -> FileResults:
    """Load cached library metadata. Rebuilds the cache if it is corrupted and
    rebuilding is allowed.

    Returns [] if cache is corrupted and rebuilding is not allowed.

    Args:
        allow_rebuild (bool, optional): Allow cache rebuilding. Defaults to True.

    Returns:
        FileResults: Cached file paths.
    """
    try:
        with open_utf8(get_library_cache_file()) as f:
            cache_data: CacheData = validate_cache_structure(json.load(f))

        results = FileResults.from_cache(cache_data)
    except (json.JSONDecodeError, KeyError):
        print_warn("Cache corrupted.")

        results = rebuild_library_cache() if allow_rebuild else []

    return results


def load_library_cache() -> FileResults:
    """Load cached library metadata. Rebuilds the cache if it has expired or is
    corrupted.

    Raises:
        typer.Exit: Cache empty or doesn't exist.

    Returns:
        FileResults: Cached file paths.
    """
    return rebuild_library_cache() if is_cache_expired() else _load_library_cache()


def is_cache_expired() -> bool:
    """Check if the library cache is older than the configured cache interval.

    Args:
        cache_timestamp (datetime): Last cache timestamp.

    Returns:
        bool: True if cache has expired or doesn't exist, False otherwise.
    """
    cache_file = get_library_cache_file()

    if not cache_file.exists():
        return True

    cache_timestamp = datetime.fromtimestamp(cache_file.stat().st_mtime)
    cache_interval: timedelta = build_config()["library_cache_interval"]

    if cache_interval.total_seconds() == 0:
        # Cache set to never expire
        return False

    return datetime.now() - cache_timestamp > cache_interval


def get_library_cache_size() -> int | None:
    """Get the size of the library cache.

    Returns:
        int | None: Number of cached file paths or None if cache doesn't exist.
    """
    return (
        len(_load_library_cache(allow_rebuild=False))
        if get_library_cache_file().exists()
        else None
    )



import os

from typer.testing import CliRunner

from mf.cli_main import app_mf

runner = CliRunner()


def _dummy_file_result(path: str):
    class DummyFile:
        def __init__(self, p):
            self._path = p
            self.name = os.path.basename(p)
            self.parent = os.path.dirname(p)

        def __str__(self):
            return self._path

    class DummyResult:
        def __init__(self, p):
            self.file = DummyFile(p)

    return DummyResult(path)


def test_play_vlc_not_found(monkeypatch):
    monkeypatch.setattr(
        "mf.cli_main.get_next", lambda: _dummy_file_result("/tmp/a.mp4")
    )
    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")
    monkeypatch.setattr(
        "mf.cli_main.get_config", lambda: {"fullscreen_playback": False}
    )

    import subprocess

    def raise_fnf(*a, **k):
        raise FileNotFoundError("vlc missing")

    monkeypatch.setattr(subprocess, "Popen", raise_fnf)

    result = runner.invoke(app_mf, ["play", "next"])
    assert result.exit_code != 0


def test_play_vlc_generic_error(monkeypatch):
    monkeypatch.setattr(
        "mf.cli_main.get_next", lambda: _dummy_file_result("/tmp/a.mp4")
    )
    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")
    monkeypatch.setattr(
        "mf.cli_main.get_config", lambda: {"fullscreen_playback": False}
    )

    import subprocess

    def raise_generic(*a, **k):
        raise Exception("boom")

    monkeypatch.setattr(subprocess, "Popen", raise_generic)

    result = runner.invoke(app_mf, ["play", "next"])
    assert result.exit_code != 0



from __future__ import annotations

import os
import shutil
import subprocess
from pathlib import Path

import typer

from ..constants import FALLBACK_EDITORS_POSIX
from .console import console, print_and_raise
from .file import FileResult


def start_editor(file: Path):
    """Open a file in an editor.

    Resolution order:
        1. VISUAL or EDITOR environment variables.
        2. Windows: Notepad++ if present else notepad.
        3. POSIX: First available editor from FALLBACK_EDITORS_POSIX.

    Args:
        file (Path): File to open.
    """
    editor = os.environ.get("VISUAL") or os.environ.get("EDITOR")
    if editor:
        subprocess.run([editor, str(file)])
        return
    if os.name == "nt":  # Windows
        if shutil.which("notepad++"):
            subprocess.run(["notepad++", str(file)])
        else:
            subprocess.run(["notepad", str(file)])
        return
    for ed in FALLBACK_EDITORS_POSIX:
        if shutil.which(ed):
            subprocess.run([ed, str(file)])
            break
    else:
        console.print(f"No editor found. Edit manually: {file}")


def open_imdb_entry(result: FileResult):
    """Print IMDB URL and open it in the default browser if one is available.

    Args:
        result (FileResult): File for which to open the IMDB entry.
    """
    # Heavy imports, import here lazily so they don't slow down every mf invocation
    from guessit import guessit
    from imdbinfo import search_title

    filestem = result.file.stem
    parsed = guessit(filestem)

    if "title" not in parsed:
        print_and_raise(f"Could not parse a title from filename '{filestem}'.")

    title = parsed["title"]
    results = search_title(title)

    if results.titles:
        imdb_url = results.titles[0].url
        console.print(f"IMDB entry for [green]{title}[/green]: {imdb_url}")
        typer.launch(imdb_url)
    else:
        print_and_raise(f"No IMDB results found for parsed title {title}.")


def format_size(size_bytes: int) -> str:
    """Format a size in bytes to human-readable string with appropriate prefix.

    Args:
        size_bytes (int): Size in bytes.

    Returns:
        str: Formatted string like "1.5 GB" or "250 MB".
    """
    units = [(1024**3, "GB"), (1024**2, "MB"), (1024**1, "kB"), (1, "B")]

    for threshold, unit in units:
        if size_bytes >= threshold:
            value = size_bytes / threshold
            # Use appropriate decimal places based on magnitude
            if value >= 100:
                return f"{value:.0f} {unit}"
            elif value >= 10:
                return f"{value:.1f} {unit}"
            else:
                return f"{value:.2f} {unit}"

    return f"{size_bytes} B"


def get_vlc_command() -> str:
    """Get the platform-specific VLC command.

    Returns:
        str: VLC command.
    """
    if os.name == "nt":
        # Try common VLC installation paths
        vlc_paths = [
            r"C:\Program Files\VideoLAN\VLC\vlc.exe",
            r"C:\Program Files (x86)\VideoLAN\VLC\vlc.exe",
        ]
        vlc_cmd = None
        for path in vlc_paths:
            if Path(path).exists():
                vlc_cmd = path
                break

        if vlc_cmd is None:
            # Try to find in PATH
            vlc_cmd = "vlc"
    else:  # Unix-like (Linux, macOS)
        vlc_cmd = "vlc"

    return vlc_cmd



import os

import pytest

from mf.utils.misc import get_vlc_command


@pytest.mark.skipif(
    os.name != "nt",
    reason="Test requires Windows (monkeypatching os.name causes Path instantiation errors on POSIX)"
)
def test_get_vlc_command_windows_prefers_known_paths():
    # Test Windows VLC path resolution
    cmd = get_vlc_command()
    # Depending on environment, it may fall back to 'vlc'
    assert cmd == "vlc" or cmd.endswith("vlc.exe")


@pytest.mark.skipif(
    os.name != "nt",
    reason="Test requires Windows (monkeypatching os.name causes Path instantiation errors on POSIX)"
)
def test_get_vlc_command_windows_falls_back_to_path():
    # Test Windows VLC fallback behavior
    cmd = get_vlc_command()
    assert cmd == "vlc" or cmd.endswith("vlc.exe")



import typer

from .utils.cache import load_library_cache, rebuild_library_cache
from .utils.config import get_config
from .utils.console import console, print_ok
from .utils.file import get_library_cache_file
from .utils.misc import format_size
from .utils.parsers import parse_resolutions
from .utils.stats import BinData, get_log_histogram, get_string_counts, show_histogram

app_cache = typer.Typer(help="Manage mf's library cache.")


@app_cache.command()
def rebuild():
    """Rebuild the library cache."""
    rebuild_library_cache()


@app_cache.command()
def file():
    """Print cache file location."""
    print(get_library_cache_file())


@app_cache.command()
def clear():
    """Clear the library cache."""
    get_library_cache_file().unlink()
    print_ok("Cleared the library cache.")


@app_cache.command()
def stats():
    """Show cache statistics."""
    cache = load_library_cache()
    media_extensions = get_config()["media_extensions"]

    if media_extensions:
        media_cache = cache.copy()
        media_cache.filter_by_extension(media_extensions)

    # Extension histogram (all files)
    console.print("")
    show_histogram(
        get_string_counts(file.suffix for file in cache.get_paths()),
        "File extensions (all files)",
        sort=True,
        # Sort by frequency descending, then name ascending
        sort_key=lambda bar: (-bar[1], bar[0]),
        top_n=20,
    )

    # Extension histogram (media file extensions only)
    if media_extensions:
        show_histogram(
            get_string_counts(file.suffix for file in media_cache.get_paths()),
            "File extensions (media files)",
            sort=True,
        )

    # Resolution distribution
    show_histogram(
        get_string_counts(parse_resolutions(cache)),
        "Media file resolution",
        sort=True,
        sort_key=lambda bar: int("".join(filter(str.isdigit, bar[0]))),
    )

    # File size distribution
    if media_extensions:
        bin_centers, bin_counts = get_log_histogram(
            [result.stat.st_size for result in media_cache]
        )

        # Centers are file sizes in bytes.
        # Convert to string with appropriate size prefix.
        bin_labels = [format_size(bin_center) for bin_center in bin_centers]

        bin_data: list[BinData] = [
            (label, count) for label, count in zip(bin_labels, bin_counts)
        ]
        show_histogram(bin_data, "Media file size")



import os

import pytest
from typer.testing import CliRunner

from mf.cli_main import app_mf

runner = CliRunner()


def test_main_callback_no_command_shows_help():
    result = runner.invoke(app_mf, [])
    # Typer Exit should occur after printing help and version
    assert result.exit_code == 0
    assert "Version:" in result.stdout
    assert "Usage:" in result.stdout or "Commands:" in result.stdout


def test_version_prints_version():
    result = runner.invoke(app_mf, ["version"])
    assert result.exit_code == 0
    # Should print semantic version string
    assert any(char.isdigit() for char in result.stdout)


def test_version_check_branch(monkeypatch):
    # Simulate check_version side effect without external calls
    called = {"value": False}

    def fake_check():
        called["value"] = True

    # cli_main imports check_version directly; patch that symbol
    monkeypatch.setattr("mf.cli_main.check_version", fake_check)
    result = runner.invoke(app_mf, ["version", "check"])
    assert result.exit_code == 0
    assert called["value"] is True


def test_find_no_results_warns(monkeypatch):
    # Force FindQuery to return empty to hit warning branch
    class Dummy:
        pattern = "*"

        def execute(self):
            return []

    monkeypatch.setattr("mf.cli_main.FindQuery", lambda p: Dummy())
    result = runner.invoke(app_mf, ["find", "*.nonexistentext"])
    assert result.exit_code == 0
    assert "No media files found" in result.stdout


def test_new_empty_collection_raises(monkeypatch):
    # Force NewQuery to return empty and hit print_and_raise path
    class Dummy:
        def __init__(self, n):
            pass

        def execute(self):
            return []

    monkeypatch.setattr("mf.cli_main.NewQuery", lambda n: Dummy(n))
    result = runner.invoke(app_mf, ["new", "5"])
    # Typer will convert raised exception to non-zero exit
    assert result.exit_code != 0
    assert "No media files found" in result.stdout or result.stderr


def test_play_invalid_target_errors(monkeypatch):
    # Pass a non-integer to trigger ValueError branch
    result = runner.invoke(app_mf, ["play", "not-an-int"])
    assert result.exit_code != 0
    assert "Invalid target" in (result.stdout + result.stderr)


def test_play_next_branch(monkeypatch):
    # Simulate get_next returning a FileResult-like object
    class DummyFile:
        name = "movie.mp4"
        parent = "/tmp"

        def __str__(self):
            return "/tmp/movie.mp4"

    class DummyResult:
        file = DummyFile()

    monkeypatch.setattr("mf.cli_main.get_next", lambda: DummyResult())
    monkeypatch.setattr("mf.cli_main.save_last_played", lambda fr: None)
    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")

    # Prevent actual process spawn
    import subprocess

    monkeypatch.setattr(subprocess, "Popen", lambda *a, **k: None)

    result = runner.invoke(app_mf, ["play", "next"])
    assert result.exit_code == 0
    assert "VLC launched successfully" in result.stdout


@pytest.mark.parametrize("fullscreen", [True, False])
def test_play_list_branch_and_fullscreen(monkeypatch, fullscreen):
    # Simulate last search results as playlist
    class DummyFile:
        def __init__(self, path):
            self._path = path
            self.name = os.path.basename(path)
            self.parent = os.path.dirname(path)

        def __str__(self):
            return self._path

    class DummyResult:
        def __init__(self, path):
            self.file = DummyFile(path)

    class DummyResults(list):
        pass

    # load_search_results returns (pattern, results, stats)
    monkeypatch.setattr(
        "mf.cli_main.load_search_results",
        lambda: (
            "pattern",
            DummyResults([DummyResult("/tmp/a.mp4"), DummyResult("/tmp/b.mp4")]),
            {},
        ),
    )

    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")
    monkeypatch.setattr(
        "mf.cli_main.get_config", lambda: {"fullscreen_playback": fullscreen}
    )

    import subprocess

    monkeypatch.setattr(subprocess, "Popen", lambda *a, **k: None)

    result = runner.invoke(app_mf, ["play", "list"])
    assert result.exit_code == 0
    assert "VLC launched successfully" in result.stdout


def test_play_random_file_branch(monkeypatch):
    # Exercise random branch by returning a deterministic list
    class DummyFile:
        def __init__(self, path):
            self._path = path
            self.name = os.path.basename(path)
            self.parent = os.path.dirname(path)

        def __str__(self):
            return self._path

    class DummyResult:
        def __init__(self, path):
            self.file = DummyFile(path)

    monkeypatch.setattr(
        "mf.cli_main.FindQuery",
        lambda p: type(
            "Q",
            (),
            {
                "execute": lambda self: [
                    DummyResult("/tmp/a.mp4"),
                    DummyResult("/tmp/b.mp4"),
                ]
            },
        )(),
    )
    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")

    import subprocess

    monkeypatch.setattr(subprocess, "Popen", lambda *a, **k: None)

    result = runner.invoke(app_mf, ["play"])
    assert result.exit_code == 0
    assert "VLC launched successfully" in result.stdout


def test_play_by_index_valid(monkeypatch):
    """Test play with valid numeric index."""
    class DummyFile:
        name = "movie.mp4"
        parent = "/tmp"

        def __str__(self):
            return "/tmp/movie.mp4"

    class DummyResult:
        file = DummyFile()

    monkeypatch.setattr("mf.cli_main.get_result_by_index", lambda i: DummyResult())
    monkeypatch.setattr("mf.cli_main.save_last_played", lambda fr: None)
    monkeypatch.setattr("mf.cli_main.get_vlc_command", lambda: "vlc")

    import subprocess

    monkeypatch.setattr(subprocess, "Popen", lambda *a, **k: None)

    result = runner.invoke(app_mf, ["play", "5"])
    assert result.exit_code == 0
    assert "VLC launched successfully" in result.stdout


def test_play_by_index_out_of_bounds(monkeypatch):
    """Test play with invalid index raises error."""
    import typer

    def raise_exit(i):
        from mf.utils.console import print_and_raise
        print_and_raise(f"Index {i} not found")

    monkeypatch.setattr("mf.cli_main.get_result_by_index", raise_exit)

    result = runner.invoke(app_mf, ["play", "999"])
    assert result.exit_code != 0


def test_play_random_empty_collection(monkeypatch):
    """Test play random when no files exist."""
    class DummyQuery:
        def execute(self):
            return []

    monkeypatch.setattr("mf.cli_main.FindQuery", lambda p: DummyQuery())

    result = runner.invoke(app_mf, ["play"])
    assert result.exit_code != 0
    assert "No media files found" in (result.stdout + result.stderr)


def test_imdb_opens(monkeypatch):
    # Ensure imdb command calls open_imdb_entry with desired index
    called = {"idx": None}

    def fake_open(fr):
        called["idx"] = True

    monkeypatch.setattr("mf.cli_main.open_imdb_entry", fake_open)
    # get_result_by_index returns a dummy FileResult-like object
    monkeypatch.setattr("mf.cli_main.get_result_by_index", lambda i: object())
    result = runner.invoke(app_mf, ["imdb", "1"])
    assert result.exit_code == 0
    assert called["idx"] is True


def test_filepath_prints(monkeypatch):
    class Dummy:
        def __init__(self, path):
            self.file = type("F", (), {"__str__": lambda self: path})()

    monkeypatch.setattr(
        "mf.cli_main.get_result_by_index", lambda i: Dummy("/tmp/x.mp4")
    )
    result = runner.invoke(app_mf, ["filepath", "1"])
    assert result.exit_code == 0
    assert "/tmp/x.mp4" in result.stdout



import subprocess
from random import randrange

import typer

from .cli_cache import app_cache
from .cli_config import app_config
from .cli_last import app_last
from .utils.config import get_config
from .utils.console import console, print_and_raise, print_warn
from .utils.file import FileResult, FileResults
from .utils.misc import get_vlc_command, open_imdb_entry
from .utils.playlist import get_next, save_last_played
from .utils.scan import FindQuery, NewQuery
from .utils.search import (
    get_result_by_index,
    load_search_results,
    print_search_results,
    save_search_results,
)
from .version import __version__, check_version

app_mf = typer.Typer(help="Media file finder and player")
app_mf.add_typer(app_last, name="last")
app_mf.add_typer(app_config, name="config")
app_mf.add_typer(app_cache, name="cache")


@app_mf.command()
def find(
    pattern: str = typer.Argument(
        "*",
        help=(
            "Search pattern (glob-based). Use quotes around patterns with wildcards "
            "to prevent shell expansion (e.g., 'mf find \"*.mp4\"'). If no wildcards "
            "are present, the pattern will be wrapped with wildcards automatically."
        ),
    ),
):
    """Find media files matching the search pattern.

    Finds matching files and prints an indexed list.
    """
    # Find, cache, and print media file paths
    query = FindQuery(pattern)
    results = query.execute()

    if not results:
        print_warn(f"No media files found matching '{query.pattern}'")
        raise typer.Exit(0)

    save_search_results(query.pattern, results)
    print_search_results(
        f"Search pattern: {query.pattern}", results, get_config()["display_paths"]
    )


@app_mf.command()
def new(
    n: int = typer.Argument(20, help="Number of latest additions to show"),
):
    """Find the latest additions to the media database."""
    newest_files = NewQuery(n).execute()
    pattern = f"{n} latest additions"

    if not newest_files:
        print_and_raise("No media files found (empty collection).")

    save_search_results(pattern, newest_files)
    print_search_results(pattern, newest_files, get_config()["display_paths"])


@app_mf.command()
def play(
    target: str = typer.Argument(
        None,
        help=(
            "Index of the file to play or 'next' to play the next search result or "
            "'list' to play last search results as a playlist. If None, plays a random "
            "file."
        ),
    ),
):
    """Play a media file by its index."""
    if target:
        if target.lower() == "next":
            file_to_play = get_next()
            save_last_played(file_to_play)
        elif target.lower() == "list":
            _, file_to_play, _ = load_search_results()
        else:
            # Play requested file
            try:
                index = int(target)
                file_to_play = get_result_by_index(index)
                save_last_played(file_to_play)
            except ValueError as e:
                print_and_raise(
                    "Invalid target: {target}. Use an index number, 'next', or 'list'.",
                    raise_from=e,
                )
    else:
        # Play random file without saving it as last played. This way a random file
        # can be played without disrupting the 'next' logic.
        all_files = FindQuery("*").execute()

        if not all_files:
            print_and_raise("No media files found (empty collection).")

        file_to_play = all_files[randrange(len(all_files))]

    # Launch VLC with the file(s)
    try:
        vlc_cmd = get_vlc_command()
        vlc_args = [vlc_cmd]

        if isinstance(file_to_play, FileResult):
            # Single file
            console.print(f"[green]Playing:[/green] {file_to_play.file.name}")
            console.print(
                f"[blue]Location:[/blue] [white]{file_to_play.file.parent}[/white]"
            )
            vlc_args.append(str(file_to_play.file))
        elif isinstance(file_to_play, FileResults):
            # Last search results as playlist
            console.print("[green]Playing:[/green] Last search results as playlist")
            vlc_args.extend(str(result.file) for result in file_to_play)

        fullscreen_playback = get_config()["fullscreen_playback"]

        if fullscreen_playback:
            vlc_args.extend(["--fullscreen", "--no-video-title-show"])

        # Launch VLC in background
        subprocess.Popen(
            vlc_args,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        console.print("[green]✓[/green] VLC launched successfully")

    except FileNotFoundError as e:
        print_and_raise("VLC not found. Please install VLC media player.", raise_from=e)

    except Exception as e:
        print_and_raise(f"Error launching VLC: {e}", raise_from=e)


@app_mf.command()
def imdb(
    index: int = typer.Argument(
        ..., help="Index of the file for which to retrieve the IMDB URL"
    ),
):
    """Open IMDB entry of a search result."""
    open_imdb_entry(get_result_by_index(index))


@app_mf.command()
def filepath(
    index: int = typer.Argument(
        ..., help="Index of the file for which to print the filepath."
    ),
):
    """Print filepath of a search result."""
    print(get_result_by_index(index).file)


@app_mf.command()
def version(
    target: str = typer.Argument(
        None,
        help="None or 'check'. If None, displays mediafinder's version. "
        "If 'check', checks if a newer version is available.",
    ),
):
    "Print version or perform version check."
    if target and target == "check":
        check_version()
    else:
        console.print(__version__)


@app_mf.callback(invoke_without_command=True)
def main_callback(ctx: typer.Context):
    """Show help when no command is provided."""
    if ctx.invoked_subcommand is None:
        console.print("")
        console.print(f" Version: {__version__}", style="bright_yellow")
        console.print(ctx.get_help())
        raise typer.Exit()



from collections.abc import Callable
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Any, Literal

from tomlkit import TOMLDocument

from ..constants import DEFAULT_MEDIA_EXTENSIONS
from .console import print_and_raise, print_ok, print_warn
from .normalizers import (
    normalize_bool_str,
    normalize_bool_to_toml,
    normalize_media_extension,
    normalize_path,
)

__all__ = [
    "apply_action",
    "REGISTRY",
    "SettingSpec",
]


def _rebuild_cache_if_enabled():
    # Helper function with lazy imports to avoid circular import
    from .cache import rebuild_library_cache
    from .config import get_config

    if get_config()["cache_library"]:
        rebuild_library_cache()


Action = Literal["set", "add", "remove", "clear"]


@dataclass
class SettingSpec:
    """Specification for a configurable setting.

    Attributes:
        key: Name of the setting in the configuration file.
        kind: Kind of setting ('scalar' or 'list').
        value_type: Python type of values loaded from TOML via from_toml.
        actions: Allowed actions for this setting.
        default: Default value(s), used in the default configuration.
        normalize: Function converting a raw string into what is written to TOML.
        from_toml: Function converting value from TOML to the typed value.
        display: Function producing a human readable representation.
        validate_all: Function validating the (possibly list) value(s).
        after_update: Hook to trigger additional action(s) after an update.
        help: Human readable help text shown to the user.
    """

    key: str
    kind: Literal["scalar", "list"]
    value_type: type
    actions: set[Action]
    default: Any
    normalize: Callable[[str], Any] = lambda value: value
    from_toml: Callable[[Any], Any] = lambda value: value
    display: Callable[[Any], str] = lambda value: str(value)
    validate_all: Callable[[Any], None] = lambda value: None
    after_update: Callable[[Any], None] = lambda value: None
    help: str = ""


REGISTRY: dict[str, SettingSpec] = {
    "search_paths": SettingSpec(
        key="search_paths",
        kind="list",
        value_type=str,
        actions={"set", "add", "remove", "clear"},
        normalize=normalize_path,
        from_toml=lambda path: Path(path).resolve(),
        default=[],
        after_update=lambda _: _rebuild_cache_if_enabled(),
        help="Directories scanned for media files.",
    ),
    "media_extensions": SettingSpec(
        key="media_extensions",
        kind="list",
        value_type=str,
        actions={"set", "add", "remove", "clear"},
        normalize=normalize_media_extension,
        default=DEFAULT_MEDIA_EXTENSIONS,
        help="Allowed media file extensions.",
    ),
    "match_extensions": SettingSpec(
        key="match_extensions",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help="Filter results by media_extensions.",
    ),
    "fullscreen_playback": SettingSpec(
        key="fullscreen_playback",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help="Play files in fullscreen mode.",
    ),
    "prefer_fd": SettingSpec(
        key="prefer_fd",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help="Use fd for file searches where possible.",
    ),
    "cache_library": SettingSpec(
        key="cache_library",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=False,
        display=normalize_bool_to_toml,
        after_update=lambda _: _rebuild_cache_if_enabled(),
        help="Cache library metadata locally.",
    ),
    "library_cache_interval": SettingSpec(
        key="library_cache_interval",
        kind="scalar",
        value_type=timedelta,
        actions={"set"},
        default=86400,
        from_toml=lambda interval_s: timedelta(seconds=int(interval_s)),
        help=(
            "Time after which the library cache is automatically rebuilt if "
            "cache_library is set to true, in seconds. Set to 0 to turn off automatic"
            "cache rebuilding. Default value of 86400 s is 1 day."
        ),
    ),
    "auto_wildcards": SettingSpec(
        key="auto_wildcards",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help=(
            "Automatically wrap search patterns with '*' if they don't contain "
            "any wildcards (* ? [ ]). 'batman' becomes '*batman*'."
        ),
    ),
    "parallel_search": SettingSpec(
        key="parallel_search",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help=(
            "Parallelize file searches over search paths. Turn off if search paths are "
            "located on the same mechanical drive (but leave on for SSD/NVME)."
        ),
    ),
    "display_paths": SettingSpec(
        key="display_paths",
        kind="scalar",
        value_type=bool,
        actions={"set"},
        normalize=normalize_bool_str,
        default=True,
        display=normalize_bool_to_toml,
        help="Display file paths in search results.",
    ),
}


def apply_action(
    cfg: TOMLDocument, key: str, action: Action, raw_values: list[str] | None
) -> TOMLDocument:
    """Apply action to setting.

    Args:
        cfg (TOMLDocument): Current configuration.
        key (str): Setting to apply action to.
        action (Action): Action to perform.
        raw_values (list[str] | None): Values to act with.

    Returns:
        TOMLDocument: Updated configuration.
    """
    if key not in REGISTRY:
        print_and_raise(
            f"Unknown configuration key: {key}. Available keys: {list(REGISTRY)}"
        )

    spec = REGISTRY[key]

    if action not in spec.actions:
        print_and_raise(f"Action {action} not supported for {key}.")

    if spec.kind == "scalar" and action == "set":
        if raw_values is None or len(raw_values) > 1:
            print_and_raise(
                f"Scalar setting {key} requires "
                f"a single value for set, got: {raw_values}."
            )

        new_value = spec.normalize(raw_values[0])
        spec.validate_all(new_value)
        cfg[key] = new_value
        spec.after_update(cfg[key])
        print_ok(f"Set {key} to '{spec.display(new_value)}'.")

        return cfg

    # List setting
    if action == "clear":
        cfg[key].clear()
        print_ok(f"Cleared {key}.")
        return cfg

    normalized_values = [spec.normalize(value) for value in raw_values]

    if action == "set":
        cfg[key].clear()
        cfg[key].extend(normalized_values)
        print_ok(f"Set {key} to {normalized_values}.")

    elif action == "add":
        for value in normalized_values:
            if value not in cfg[key]:
                cfg[key].append(value)
                print_ok(f"Added '{value}' to {key}.")
            else:
                print_warn(f"{key} already contains '{value}', skipping.")

    elif action == "remove":
        for value in normalized_values:
            if value in cfg[key]:
                cfg[key].remove(value)
                print_ok(f"Removed '{value}' from {key}.")
            else:
                print_warn(f"'{value}' not found in {key}, skipping.")

    spec.validate_all(cfg[key])
    spec.after_update(cfg[key])

    return cfg
